import re
from typing import List, Union

from spacy.tokens import Span

from ..definitions import Entity, Replacement
from ..constants import ENTITY_TYPES

# =====================================
# Entity converters
# =====================================


def convert_spacy_to_entity(
    entity: Span,
    type: ENTITY_TYPES = None,
    regex: Union[str, re.Pattern] = ".*",
    *args,
    **kwargs,
) -> Entity:
    """Convert a spacy entity to an anonipy entity object.

    Args:
        entity: The spacy Span representing the entity to convert.
        type: The type of the entity.
        regex: The regular expression the entity must match.

    Returns:
        The converted anonipy entity object.

    """

    return Entity(
        entity.text,
        entity.label_,
        entity.start_char,
        entity.end_char,
        entity._.score,
        type,
        regex,
    )


# =====================================
# Anonymization function
# =====================================


def anonymize(text: str, replacements: List[Replacement]) -> str:
    """Anonymize a text based on a list of replacements.

    Examples:
        >>> from anonipy.anonymize import anonymize
        >>> anonymize(text, replacements)

    Args:
        text: The text to anonymize.
        replacements: The list of replacements to apply.

    Returns:
        The anonymized text.

    """

    s_replacements = sorted(replacements, key=lambda x: x["start_index"], reverse=True)

    anonymized_text = text
    for replacement in s_replacements:
        anonymized_text = (
            anonymized_text[: replacement["start_index"]]
            + replacement["anonymized_text"]
            + anonymized_text[replacement["end_index"] :]
        )
    return anonymized_text, s_replacements[::-1]



"""Module containing the `regex` related utilities.

The `regex` module provides the regex definitions and functions used within the package.

Classes:
    RegexMapping: The class representing the mapping for data type to the corresponding regex.

Attributes:
    REGEX_STRING (str): The regex definition for string.
    REGEX_INTEGER (str): The regex definition for integer.
    REGEX_FLOAT (str): The regex definition for float.
    REGEX_DATE (str): The regex definition for date.
    REGEX_EMAIL_ADDRESS (str): The regex definition for email address.
    REGEX_PHONE_NUMBER (str): The regex definition for phone number.
    REGEX_WEBSITE_URL (str): The regex definition for website URL.

"""

from collections import defaultdict

from ..constants import ENTITY_TYPES


# =====================================
# Regex definitions
# =====================================

REGEX_STRING = ".*"
"""The regex definition for string."""

REGEX_INTEGER = "\d+"
"""The regex definition for integer."""

REGEX_FLOAT = "[\d\.,]+"
"""The regex definition for float."""

REGEX_DATE = (
    r"("
    r"(\d{4}[-/.\s]\d{2}[-/.\s]\d{2}[ T]\d{2}:\d{2}:\d{2})|"
    r"(\d{2}[-/.\s]\d{2}[-/.\s]\d{4}[ T]\d{2}:\d{2}:\d{2})|"
    r"(\d{2}[-/.\s]\d{2}[-/.\s]\d{4}[ T]\d{2}:\d{2})|"
    r"(\d{4}[-/.\s]\d{2}[-/.\s]\d{2}[ T]\d{2}:\d{2})|"
    r"(\d{4}[-/.\s]\d{2}[-/.\s]\d{2}[ T]\d{2}:\d{2} [APap][mM])|"
    r"(\d{2}[-/.\s]\d{2}[-/.\s]\d{4}[ T]\d{2}:\d{2} [APap][mM])|"
    r"(\d{4}[-/.\s]\d{2}[-/.\s]\d{2})|"
    r"(\d{2}[-/.\s]\d{2}[-/.\s]\d{4})|"
    r"(\d{2}[-/.\s]\d{2}[-/.\s]\d{4}[ ]?\d{2}:\d{2}:\d{2})|"
    r"(\d{4}[-/.\s]\d{2}[-/.\s]\d{2}[ ]?\d{2}:\d{2}:\d{2})|"
    r"(\d{1,2}[ ](January|February|March|April|May|June|July|August|September|October|November|December)[ ]\d{4}[ ]?\d{2}:\d{2}:\d{2})|"
    r"(\d{1,2}[ ](Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[ ]\d{4}[ ]?\d{2}:\d{2}:\d{2})|"
    r"(\d{1,2}[ ](January|February|March|April|May|June|July|August|September|October|November|December)[ ]\d{4}[ ]?\d{2}:\d{2}[ ]?[APap][mM])|"
    r"(\d{1,2}[ ](Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[ ]\d{4}[ ]?\d{2}:\d{2}[ ]?[APap][mM])|"
    r"([A-Za-z]+,[ ]\d{1,2}[ ](January|February|March|April|May|June|July|August|September|October|November|December)[ ]\d{4}[ ]?\d{2}:\d{2}:\d{2})|"
    r"([A-Za-z]+,[ ](January|February|March|April|May|June|July|August|September|October|November|December)[ ]\d{1,2},[ ]\d{4}[ ]?\d{2}:\d{2}:\d{2})|"
    r"([A-Za-z]+,[ ]\d{1,2}[ ](January|February|March|April|May|June|July|August|September|October|November|December)[ ]\d{4}[ ]?\d{2}:\d{2}[ ]?[APap][mM])|"
    r"([A-Za-z]+,[ ](January|February|March|April|May|June|July|August|September|October|November|December)[ ]\d{1,2},[ ]\d{4}[ ]?\d{2}:\d{2}[ ]?[APap][mM])"
    r")"
)
"""The regex definition for dates.

The regex definition for dates includes string representations, which are currently in
the English language.

TODO:
    - Add regex definitions for other languages.
"""

REGEX_EMAIL_ADDRESS = (
    "[a-zA-Z0-9.!#$%&’*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*"
)
"""The regex definition for email addresses."""

REGEX_PHONE_NUMBER = (
    "[(]?[\+]?[(]?[0-9]{1,3}[)]?[-\s\.]?([0-9]{2,}[-\s\.]?){2,}([0-9]{3,})"
)
"""The regex definition for phone numbers."""

REGEX_WEBSITE_URL = "((https?|ftp|smtp):\/\/)?(www.)?([a-zA-Z0-9]+\.)+[a-z]{2,}(\/[a-zA-Z0-9#\?\_\.\=\-\&]+|\/?)*"
"""The regex definition for website URLs."""


# =====================================
# Define the regex definitions
# =====================================


class RegexMapping:
    """The class representing the regex mapping.

    Examples:
        >>> from anonipy.anonymize.regex import regex_mapping
        >>> regex_mapping["string"]
        ".*"

    Attributes:
        regex_mapping (defaultdict):
            The mapping between the data type and the corresponding regex.

    Methods:
        __getitem__(type):
            Gets the regex for the given type.

    """

    def __init__(self):
        """Initialize the regex mapping.

        Examples:
            >>> from anonipy.anonymize.regex import RegexMapping
            >>> regex_mapping = RegexMapping()

        """

        self.regex_mapping = defaultdict(lambda: ".*")
        # Define the regex mappings
        self.regex_mapping[ENTITY_TYPES.STRING] = REGEX_STRING
        self.regex_mapping[ENTITY_TYPES.INTEGER] = REGEX_INTEGER
        self.regex_mapping[ENTITY_TYPES.FLOAT] = REGEX_FLOAT
        self.regex_mapping[ENTITY_TYPES.DATE] = REGEX_DATE
        self.regex_mapping[ENTITY_TYPES.EMAIL] = REGEX_EMAIL_ADDRESS
        self.regex_mapping[ENTITY_TYPES.PHONE_NUMBER] = REGEX_PHONE_NUMBER
        self.regex_mapping[ENTITY_TYPES.WEBSITE_URL] = REGEX_WEBSITE_URL

    def __getitem__(self, type: str) -> str:
        """Gets the regex for the given type.

        Examples:
            >>> from anonipy.anonymize.regex import RegexMapping
            >>> regex_mapping = RegexMapping()
            >>> regex_mapping["string"]
            ".*"

        Args:
            type: The type of the entity.

        Returns:
            The regex for the given type.

        """
        return self.regex_mapping[type]


regex_mapping = RegexMapping()
"""The shorthand to the `RegexMapping` instance."""



from typing import List, Set, Tuple, Iterable

import itertools

from spacy import displacy
from spacy.tokens import Doc

from ...definitions import Entity
from ...utils.colors import get_label_color

from .interface import ExtractorInterface


# ===============================================
# Extractor class
# ===============================================


class MultiExtractor:
    """The class representing the multi extractor.

    Examples:
        >>> from anonipy.constants import LANGUAGES
        >>> from anonipy.anonymize.extractors import NERExtractor, PatternExtractor, MultiExtractor
        >>> extractors = [
        >>>     NERExtractor(ner_labels, lang=LANGUAGES.ENGLISH),
        >>>     PatternExtractor(pattern_labels, lang=LANGUAGES.ENGLISH),
        >>> ]
        >>> extractor = MultiExtractor(extractors)
        >>> extractor("John Doe is a 19 year old software engineer.")
        [(Doc, [Entity]), (Doc, [Entity])], [Entity]

    Attributes:
        extractors (List[ExtractorInterface]):
            The list of extractors to use.

    Methods:
        __call__(self, text):
            Extract the entities fron the text using the provided extractors.
        display(self, doc):
            Display the entities extracted from the text document.

    """

    def __init__(self, extractors: List[ExtractorInterface]):
        """Initialize the multi extractor.

        Examples:
            >>> from anonipy.constants import LANGUAGES
            >>> from anonipy.anonymize.extractors import NERExtractor, PatternExtractor, MultiExtractor
            >>> extractors = [
            >>>     NERExtractor(ner_labels, lang=LANGUAGES.ENGLISH),
            >>>     PatternExtractor(pattern_labels, lang=LANGUAGES.ENGLISH),
            >>> ]
            >>> extractor = MultiExtractor(extractors)
            MultiExtractor()

        Args:
            extractors: The list of extractors to use.

        """

        self.extractors = extractors

    def __call__(
        self, text: str
    ) -> Tuple[List[Tuple[Doc, List[Entity]]], List[Entity]]:
        """Extract the entities fron the text using the provided extractors.

        Examples:
            >>> extractor("John Doe is a 19 year old software engineer.")
            [(Doc, [Entity]), (Doc, [Entity])], [Entity]

        Args:
            text: The text to extract entities from.

        Returns:
            The list of extractor outputs containing the tuple (spacy document, extracted entities).
            The list of joint entities.
        """

        extractor_outputs = [e(text) for e in self.extractors]
        joint_entities = self._merge_entities(extractor_outputs)
        return extractor_outputs, joint_entities

    def display(self, doc: Doc, page: bool = False, jupyter: bool = None) -> str:
        """Display the entities in the text.

        Examples:
            >>> extractor_outputs, entities = extractor("John Doe is a 19 year old software engineer.")
            >>> extractor.display(extractor_outputs[0][0])
            HTML

        Args:
            doc: The spacy doc to display.
            page: Whether to display the doc in a web browser.
            jupyter: Whether to display the doc in a jupyter notebook.

        Returns:
            The HTML representation of the document and the extracted entities.

        """

        labels = list(
            itertools.chain.from_iterable([e.labels for e in self.extractors])
        )
        options = {"colors": {l["label"]: get_label_color(l["label"]) for l in labels}}
        return displacy.render(
            doc, style="ent", options=options, page=page, jupyter=jupyter
        )

    def _merge_entities(
        self, extractor_outputs: List[Tuple[Doc, List[Entity]]]
    ) -> List[Entity]:
        """Merges the entities returned by the extractors.

        Args:
            extractor_outputs: The list of extractor outputs.

        Returns:
            The merged entities list.

        """

        if len(extractor_outputs) == 0:
            return []
        if len(extractor_outputs) == 1:
            return extractor_outputs[1]

        joint_entities = self._filter_entities(
            list(
                itertools.chain.from_iterable(
                    [entity[1] for entity in extractor_outputs]
                )
            )
        )
        return joint_entities

    def _filter_entities(self, entities: Iterable[Entity]) -> List[Entity]:
        """Filters the entities based on their start and end indices.

        Args:
            entities: The entities to filter.

        Returns:
            The filtered entities.

        """

        get_sort_key = lambda entity: (
            entity.end_index - entity.start_index,
            -entity.start_index,
        )
        sorted_entities = sorted(entities, key=get_sort_key, reverse=True)
        result = []
        seen_tokens: Set[int] = set()
        for entities in sorted_entities:
            # Check for end - 1 here because boundaries are inclusive
            if (
                entities.start_index not in seen_tokens
                and entities.end_index - 1 not in seen_tokens
            ):
                result.append(entities)
                seen_tokens.update(range(entities.start_index, entities.end_index))
        result = sorted(result, key=lambda entity: entity.start_index)
        return result



import unittest
import warnings

import torch

from anonipy.definitions import Entity
from anonipy.anonymize.extractors import NERExtractor, PatternExtractor, MultiExtractor
from anonipy.utils.regex import regex_mapping
from anonipy.constants import LANGUAGES


# =====================================
# Helper functions
# =====================================

original_text = """\
Medical Record

Patient Name: John Doe
Date of Birth: 15-01-1985
Date of Examination: 20-05-2024
Social Security Number: 123-45-6789

Examination Procedure:
John Doe underwent a routine physical examination. The procedure included measuring vital signs (blood pressure, heart rate, temperature), a comprehensive blood panel, and a cardiovascular stress test. The patient also reported occasional headaches and dizziness, prompting a neurological assessment and an MRI scan to rule out any underlying issues.

Medication Prescribed:

Ibuprofen 200 mg: Take one tablet every 6-8 hours as needed for headache and pain relief.
Lisinopril 10 mg: Take one tablet daily to manage high blood pressure.
Next Examination Date:
15-11-2024
"""

ner_entities = [
    Entity(
        text="John Doe",
        label="name",
        start_index=30,
        end_index=38,
        type="string",
    ),
    Entity(
        text="15-01-1985",
        label="date of birth",
        start_index=54,
        end_index=64,
        type="date",
    ),
    Entity(
        text="20-05-2024",
        label="date",
        start_index=86,
        end_index=96,
        type="date",
    ),
    Entity(
        text="123-45-6789",
        label="social security number",
        start_index=121,
        end_index=132,
        type="custom",
        regex="[0-9]{3}-[0-9]{2}-[0-9]{4}",
    ),
    Entity(
        text="John Doe",
        label="name",
        start_index=157,
        end_index=165,
        type="string",
    ),
    Entity(
        text="15-11-2024",
        label="date",
        start_index=717,
        end_index=727,
        type="date",
    ),
]

pattern_entities = [
    Entity(
        text="15-01-1985",
        label="date",
        start_index=54,
        end_index=64,
        type=None,
    ),
    Entity(
        text="20-05-2024",
        label="date",
        start_index=86,
        end_index=96,
        type=None,
    ),
    Entity(
        text="blood pressure, heart rate, temperature",
        label="symptoms",
        start_index=254,
        end_index=293,
        type=None,
        regex="\\((.*)\\)",
    ),
    Entity(
        text="Ibuprofen 200 mg",
        label="medicine",
        start_index=533,
        end_index=549,
        type=None,
    ),
    Entity(
        text="Lisinopril 10 mg",
        label="medicine",
        start_index=623,
        end_index=639,
        type=None,
    ),
    Entity(
        text="15-11-2024",
        label="date",
        start_index=717,
        end_index=727,
        type=None,
    ),
]


# =====================================
# Test NER Extractor
# =====================================


class TestNERExtractor(unittest.TestCase):

    def setUp(self):
        warnings.filterwarnings("ignore", category=ImportWarning)
        warnings.filterwarnings("ignore", category=UserWarning)
        warnings.filterwarnings("ignore", category=FutureWarning)
        # define the labels to be extracted and anonymized
        self.labels = [
            {"label": "name", "type": "string"},
            {
                "label": "social security number",
                "type": "custom",
                "regex": "[0-9]{3}-[0-9]{2}-[0-9]{4}",
            },
            {"label": "date of birth", "type": "date"},
            {"label": "date", "type": "date"},
        ]

    def test_init(self):
        try:
            NERExtractor()
        except Exception as e:
            self.assertRaises(TypeError, e)

    def test_init_inputs(self):
        extractor = NERExtractor(
            labels=self.labels, lang=LANGUAGES.ENGLISH, score_th=0.5
        )
        self.assertEqual(extractor.__class__, NERExtractor)

    def test_init_gpu(self):
        if torch.cuda.is_available():
            extractor = NERExtractor(
                labels=self.labels, lang=LANGUAGES.ENGLISH, score_th=0.5, use_gpu=True
            )
            self.assertEqual(extractor.__class__, NERExtractor)

    def test_methods(self):
        extractor = NERExtractor(
            labels=self.labels, lang=LANGUAGES.ENGLISH, score_th=0.5
        )
        self.assertEqual(hasattr(extractor, "__call__"), True)
        self.assertEqual(hasattr(extractor, "display"), True)

    def test_extract_default(self):
        extractor = NERExtractor(
            labels=self.labels, lang=LANGUAGES.ENGLISH, score_th=0.5
        )
        doc, entities = extractor(original_text)
        for p_entity, t_entity in zip(entities, ner_entities):
            self.assertEqual(p_entity.text, t_entity.text)
            self.assertEqual(p_entity.label, t_entity.label)
            self.assertEqual(p_entity.start_index, t_entity.start_index)
            self.assertEqual(p_entity.end_index, t_entity.end_index)
            self.assertEqual(p_entity.type, t_entity.type)
            self.assertEqual(p_entity.regex, t_entity.regex)
            self.assertEqual(p_entity.score >= 0.5, True)


# =====================================
# Test Pattern Extractor
# =====================================


class TestPatternExtractor(unittest.TestCase):

    def setUp(self):
        warnings.filterwarnings("ignore", category=ImportWarning)
        warnings.filterwarnings("ignore", category=UserWarning)
        warnings.filterwarnings("ignore", category=FutureWarning)
        # define the labels to be extracted and anonymized
        self.labels = [
            {
                "label": "symptoms",
                "regex": r"\((.*)\)",  # symptoms are enclosed in parentheses
            },
            {
                "label": "medicine",
                "pattern": [[{"IS_ALPHA": True}, {"LIKE_NUM": True}, {"LOWER": "mg"}]],
            },
            {
                "label": "date",
                "pattern": [  # represent the date as a sequence of digits using spacy
                    [
                        {"SHAPE": "dd"},
                        {"TEXT": "-"},
                        {"SHAPE": "dd"},
                        {"TEXT": "-"},
                        {"SHAPE": "dddd"},
                    ]
                ],
            },
        ]

    def test_init(self):
        try:
            PatternExtractor()
        except Exception as e:
            self.assertRaises(TypeError, e)

    def test_init_inputs(self):
        extractor = PatternExtractor(labels=self.labels, lang=LANGUAGES.ENGLISH)
        self.assertEqual(extractor.__class__, PatternExtractor)

    def test_methods(self):
        extractor = PatternExtractor(labels=self.labels, lang=LANGUAGES.ENGLISH)
        self.assertEqual(hasattr(extractor, "__call__"), True)
        self.assertEqual(hasattr(extractor, "display"), True)

    def test_extract_default(self):
        extractor = PatternExtractor(labels=self.labels, lang=LANGUAGES.ENGLISH)
        doc, entities = extractor(original_text)
        for p_entity, t_entity in zip(entities, pattern_entities):
            self.assertEqual(p_entity.text, t_entity.text)
            self.assertEqual(p_entity.label, t_entity.label)
            self.assertEqual(p_entity.start_index, t_entity.start_index)
            self.assertEqual(p_entity.end_index, t_entity.end_index)
            self.assertEqual(p_entity.type, t_entity.type)
            self.assertEqual(p_entity.regex, t_entity.regex)
            self.assertEqual(p_entity.score == 1.0, True)


class TestMultiExtractor(unittest.TestCase):

    def setUp(self):
        warnings.filterwarnings("ignore", category=ImportWarning)
        warnings.filterwarnings("ignore", category=UserWarning)
        warnings.filterwarnings("ignore", category=FutureWarning)
        # define the labels to be extracted and anonymized
        self.ner_labels = [
            {"label": "name", "type": "string"},
            {
                "label": "social security number",
                "type": "custom",
                "regex": "[0-9]{3}-[0-9]{2}-[0-9]{4}",
            },
            {"label": "date of birth", "type": "date"},
            {"label": "date", "type": "date"},
        ]
        self.pattern_labels = [
            {
                "label": "symptoms",
                "regex": r"\((.*)\)",  # symptoms are enclosed in parentheses
            },
            {
                "label": "medicine",
                "pattern": [[{"IS_ALPHA": True}, {"LIKE_NUM": True}, {"LOWER": "mg"}]],
            },
            {
                "label": "date",
                "pattern": [  # represent the date as a sequence of digits using spacy
                    [
                        {"SHAPE": "dd"},
                        {"TEXT": "-"},
                        {"SHAPE": "dd"},
                        {"TEXT": "-"},
                        {"SHAPE": "dddd"},
                    ]
                ],
            },
        ]

    def test_init(self):
        try:
            MultiExtractor()
        except Exception as e:
            self.assertRaises(TypeError, e)

    def test_init_inputs(self):
        extractors = [
            NERExtractor(labels=self.ner_labels, lang=LANGUAGES.ENGLISH),
            PatternExtractor(labels=self.pattern_labels, lang=LANGUAGES.ENGLISH),
        ]
        extractor = MultiExtractor(extractors)
        self.assertEqual(extractor.__class__, MultiExtractor)

    def test_methods(self):
        extractors = [
            NERExtractor(labels=self.ner_labels, lang=LANGUAGES.ENGLISH),
            PatternExtractor(labels=self.pattern_labels, lang=LANGUAGES.ENGLISH),
        ]
        extractor = MultiExtractor(extractors)
        self.assertEqual(hasattr(extractor, "__call__"), True)
        self.assertEqual(hasattr(extractor, "display"), True)

    def test_extract_default(self):
        extractors = [
            NERExtractor(labels=self.ner_labels, lang=LANGUAGES.ENGLISH),
            PatternExtractor(labels=self.pattern_labels, lang=LANGUAGES.ENGLISH),
        ]
        extractor = MultiExtractor(extractors)
        extractor_outputs, joint_entities = extractor(original_text)

        # check the performance of the first extractor
        for p_entity, t_entity in zip(extractor_outputs[0][1], ner_entities):
            self.assertEqual(p_entity.text, t_entity.text)
            self.assertEqual(p_entity.label, t_entity.label)
            self.assertEqual(p_entity.start_index, t_entity.start_index)
            self.assertEqual(p_entity.end_index, t_entity.end_index)
            self.assertEqual(p_entity.type, t_entity.type)
            self.assertEqual(p_entity.regex, t_entity.regex)
            self.assertEqual(p_entity.score >= 0.5, True)

        # check the performance of the second extractor
        for p_entity, t_entity in zip(extractor_outputs[1][1], pattern_entities):
            self.assertEqual(p_entity.text, t_entity.text)
            self.assertEqual(p_entity.label, t_entity.label)
            self.assertEqual(p_entity.start_index, t_entity.start_index)
            self.assertEqual(p_entity.end_index, t_entity.end_index)
            self.assertEqual(p_entity.type, t_entity.type)
            self.assertEqual(p_entity.regex, t_entity.regex)
            self.assertEqual(p_entity.score == 1.0, True)

        # check the performance of the joint entities generation
        for p_entity, t_entity in zip(
            joint_entities, extractor._filter_entities(ner_entities + pattern_entities)
        ):
            self.assertEqual(p_entity.text, t_entity.text)
            self.assertEqual(p_entity.label, t_entity.label)
            self.assertEqual(p_entity.start_index, t_entity.start_index)
            self.assertEqual(p_entity.end_index, t_entity.end_index)
            self.assertEqual(p_entity.type, t_entity.type)
            self.assertEqual(p_entity.regex, t_entity.regex)
            self.assertEqual(p_entity.score >= 0.5, True)


if __name__ == "__main__":
    unittest.main()
